# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set, Union, cast

import numpy as np
import pandas as pd
from openpyxl import load_workbook

from taipy.common.config.common.scope import Scope

from .._entity._reload import _Reloader
from .._version._version_manager_factory import _VersionManagerFactory
from ..exceptions.exceptions import ExposedTypeLengthMismatch, NonExistingExcelSheet, SheetNameLengthMismatch
from ..job.job_id import JobId
from ._file_datanode_mixin import _FileDataNodeMixin
from ._tabular_datanode_mixin import _TabularDataNodeMixin
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit


class ExcelDataNode(DataNode, _FileDataNodeMixin, _TabularDataNodeMixin):
    """Data Node stored as an Excel file.

    The Excel file format is _xlsx_.

    The *properties* attribute can contain the following optional entries:

    - *sheet_name* (`Union[str, List[str]]`): The name of the sheet(s) to be used.
    - *default_path* (`str`): The default path of the Excel file used at the instantiation of the
        data node.
    - *default_data*: The default data of the data node. It is used at the data node instantiation
        to write the data to the Excel file.
    - *has_header* (`bool`): If True, indicates that the Excel file has a header.
    - *exposed_type* (`str`): The exposed type of the data read from Excel file. The default value
        is `pandas`.
    """

    __STORAGE_TYPE = "excel"
    __SHEET_NAME_PROPERTY = "sheet_name"

    _REQUIRED_PROPERTIES: List[str] = []

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: List[Edit] = None,
        version: str = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        editor_id: Optional[str] = None,
        editor_expiration_date: Optional[datetime] = None,
        properties: Dict = None,
    ) -> None:
        self.id = id or self._new_id(config_id)

        if properties is None:
            properties = {}

        if self.__SHEET_NAME_PROPERTY not in properties.keys():
            properties[self.__SHEET_NAME_PROPERTY] = None
        if self._HAS_HEADER_PROPERTY not in properties.keys():
            properties[self._HAS_HEADER_PROPERTY] = True
        properties[self._EXPOSED_TYPE_PROPERTY] = _TabularDataNodeMixin._get_valid_exposed_type(properties)
        self._check_exposed_type(properties[self._EXPOSED_TYPE_PROPERTY])

        default_value = properties.pop(self._DEFAULT_DATA_KEY, None)
        _FileDataNodeMixin.__init__(self, properties)
        _TabularDataNodeMixin.__init__(self, **properties)

        DataNode.__init__(
            self,
            config_id,
            scope,
            self.id,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            editor_id,
            editor_expiration_date,
            **properties,
        )

        with _Reloader():
            self._write_default_data(default_value)

        self._TAIPY_PROPERTIES.update(
            {
                self._PATH_KEY,
                self._DEFAULT_PATH_KEY,
                self._DEFAULT_DATA_KEY,
                self._IS_GENERATED_KEY,
                self._HAS_HEADER_PROPERTY,
                self._EXPOSED_TYPE_PROPERTY,
                self.__SHEET_NAME_PROPERTY,
            }
        )

    @classmethod
    def storage_type(cls) -> str:
        """Return the storage type of the data node: "excel"."""
        return cls.__STORAGE_TYPE

    def write_with_column_names(self, data: Any, columns: List[str] = None, job_id: Optional[JobId] = None) -> None:
        """Write a set of columns.

        Parameters:
            data (Any): The data to write.
            columns (List[str]): The list of column names to write.
            job_id (Optional[JobId]): An optional identifier of the writer.
        """
        if isinstance(data, Dict) and all(isinstance(x, (pd.DataFrame, np.ndarray)) for x in data.values()):
            self._write_excel_with_multiple_sheets(data, columns=columns)
        else:
            df = pd.DataFrame(data)
            if columns:
                df = self._set_column_if_dataframe(df, columns)
            self._write_excel_with_single_sheet(df.to_excel, self.path, index=False)
        self.track_edit(timestamp=datetime.now(), job_id=job_id)

    @staticmethod
    def _check_exposed_type(exposed_type):
        if isinstance(exposed_type, str):
            _TabularDataNodeMixin._check_exposed_type(exposed_type)
        elif isinstance(exposed_type, list):
            for t in exposed_type:
                _TabularDataNodeMixin._check_exposed_type(t)
        elif isinstance(exposed_type, dict):
            for t in exposed_type.values():
                _TabularDataNodeMixin._check_exposed_type(t)

    def _read(self):
        return self._read_from_path()

    def _read_from_path(self, path: Optional[str] = None, **read_kwargs) -> Any:
        if path is None:
            path = self._path

        exposed_type = self.properties[self._EXPOSED_TYPE_PROPERTY]
        if exposed_type == self._EXPOSED_TYPE_PANDAS:
            return self._read_as_pandas_dataframe(path=path)
        if exposed_type == self._EXPOSED_TYPE_NUMPY:
            return self._read_as_numpy(path=path)
        return self._read_as(path=path)

    def _read_sheet_with_exposed_type(
        self, path: str, sheet_exposed_type: str, sheet_name: str
    ) -> Optional[Union[np.ndarray, pd.DataFrame]]:
        if sheet_exposed_type == self._EXPOSED_TYPE_NUMPY:
            return self._read_as_numpy(path, sheet_name)
        elif sheet_exposed_type == self._EXPOSED_TYPE_PANDAS:
            return self._read_as_pandas_dataframe(path, sheet_name)  # type: ignore
        return None

    def _read_as(self, path: str):
        try:
            properties = self.properties
            excel_file = load_workbook(path, read_only=True)
            exposed_type = properties[self._EXPOSED_TYPE_PROPERTY]
            work_books = {}
            sheet_names = excel_file.sheetnames

            user_provided_sheet_names = properties.get(self.__SHEET_NAME_PROPERTY) or []
            if not isinstance(user_provided_sheet_names, (list, set, tuple)):
                user_provided_sheet_names = [user_provided_sheet_names]

            provided_sheet_names = user_provided_sheet_names or sheet_names

            for sheet_name in provided_sheet_names:
                if sheet_name not in sheet_names:
                    raise NonExistingExcelSheet(sheet_name, path)

            if isinstance(exposed_type, List):
                if len(provided_sheet_names) != len(exposed_type):
                    raise ExposedTypeLengthMismatch(
                        f"Expected {len(provided_sheet_names)} exposed types, got " f"{len(exposed_type)}"
                    )

            for i, sheet_name in enumerate(provided_sheet_names):
                work_sheet = excel_file[sheet_name]
                sheet_exposed_type = exposed_type

                if not isinstance(sheet_exposed_type, str):
                    if isinstance(exposed_type, dict):
                        sheet_exposed_type = exposed_type.get(sheet_name, self._EXPOSED_TYPE_PANDAS)
                    elif isinstance(exposed_type, List):
                        sheet_exposed_type = exposed_type[i]

                    if isinstance(sheet_exposed_type, str):
                        sheet_data = self._read_sheet_with_exposed_type(path, sheet_exposed_type, sheet_name)
                        if sheet_data is not None:
                            work_books[sheet_name] = sheet_data
                        continue

                res = [[col.value for col in row] for row in work_sheet.rows]
                if properties[self._HAS_HEADER_PROPERTY] and res:
                    header = res.pop(0)
                    for i, row in enumerate(res):
                        res[i] = sheet_exposed_type(**dict([[h, r] for h, r in zip(header, row)]))
                else:
                    for i, row in enumerate(res):
                        res[i] = sheet_exposed_type(*row)
                work_books[sheet_name] = res  # type: ignore
        finally:
            excel_file.close()

        if len(user_provided_sheet_names) == 1:
            return work_books[cast(list, user_provided_sheet_names)[0]]

        return work_books

    def _read_as_numpy(self, path: str, sheet_names=None):
        sheets = self._read_as_pandas_dataframe(path=path, sheet_names=sheet_names)
        if isinstance(sheets, dict):
            return {sheet_name: df.to_numpy() for sheet_name, df in sheets.items()}
        return sheets.to_numpy()

    def _do_read_excel(
        self, path: str, sheet_names, kwargs
    ) -> Union[Dict[Union[int, str], pd.DataFrame], pd.DataFrame]:
        return pd.read_excel(path, sheet_name=sheet_names, **kwargs)

    def __get_sheet_names_and_header(self, sheet_names):
        kwargs = {}
        properties = self.properties
        if sheet_names is None:
            sheet_names = properties[self.__SHEET_NAME_PROPERTY]
        if not properties[self._HAS_HEADER_PROPERTY]:
            kwargs["header"] = None
        return sheet_names, kwargs

    def _read_as_pandas_dataframe(
        self, path: str, sheet_names=None
    ) -> Union[Dict[Union[int, str], pd.DataFrame], pd.DataFrame]:
        sheet_names, kwargs = self.__get_sheet_names_and_header(sheet_names)
        try:
            return self._do_read_excel(path, sheet_names, kwargs)
        except pd.errors.EmptyDataError:
            return pd.DataFrame()

    def _append_excel_with_single_sheet(self, append_excel_fct, *args, **kwargs):
        sheet_name = self.properties.get(self.__SHEET_NAME_PROPERTY)

        with pd.ExcelWriter(self._path, mode="a", engine="openpyxl", if_sheet_exists="overlay") as writer:
            if sheet_name:
                if not isinstance(sheet_name, str):
                    sheet_name = sheet_name[0]
                append_excel_fct(
                    writer, *args, **kwargs, sheet_name=sheet_name, startrow=writer.sheets[sheet_name].max_row
                )
            else:
                sheet_name = list(writer.sheets.keys())[0]
                append_excel_fct(writer, *args, **kwargs, startrow=writer.sheets[sheet_name].max_row)

    def _set_column_if_dataframe(self, data: Any, columns) -> Union[pd.DataFrame, Any]:
        if isinstance(data, pd.DataFrame):
            data.columns = pd.Index(columns, dtype="object")
        return data

    def _append_excel_with_multiple_sheets(self, data: Any, columns: List[str] = None):
        with pd.ExcelWriter(self._path, mode="a", engine="openpyxl", if_sheet_exists="overlay") as writer:
            # Each key stands for a sheet name
            for sheet_name in data.keys():
                if isinstance(data[sheet_name], np.ndarray):
                    df = pd.DataFrame(data[sheet_name])
                else:
                    df = data[sheet_name]

                if columns:
                    df = self._set_column_if_dataframe(df, columns)

                df.to_excel(
                    writer, sheet_name=sheet_name, index=False, header=False, startrow=writer.sheets[sheet_name].max_row
                )

    def _append(self, data: Any):
        from importlib.metadata import version

        if version("pandas") < "1.4":
            raise ImportError("The append method is only available for pandas version 1.4 or higher.")

        if isinstance(data, Dict) and all(isinstance(x, (pd.DataFrame, np.ndarray)) for x in data.values()):
            self._append_excel_with_multiple_sheets(data)
        elif isinstance(data, pd.DataFrame):
            self._append_excel_with_single_sheet(data.to_excel, index=False, header=False)
        else:
            self._append_excel_with_single_sheet(pd.DataFrame(data).to_excel, index=False, header=False)

    def _write_excel_with_single_sheet(self, write_excel_fct, *args, **kwargs):
        if sheet_name := self.properties.get(self.__SHEET_NAME_PROPERTY):
            if not isinstance(sheet_name, str):
                if len(sheet_name) > 1:
                    raise SheetNameLengthMismatch
                else:
                    sheet_name = sheet_name[0]
            write_excel_fct(*args, **kwargs, sheet_name=sheet_name)
        else:
            write_excel_fct(*args, **kwargs)

    def _write_excel_with_multiple_sheets(self, data: Any, columns: List[str] = None):
        with pd.ExcelWriter(self._path) as writer:
            # Each key stands for a sheet name
            properties = self.properties
            for key in data.keys():
                df = self._convert_data_to_dataframe(properties[self._EXPOSED_TYPE_PROPERTY], data[key])

                if columns:
                    df = self._set_column_if_dataframe(df, columns)

                df.to_excel(writer, key, index=False, header=properties[self._HAS_HEADER_PROPERTY] or False)

    def _write(self, data: Any):
        if isinstance(data, Dict):
            return self._write_excel_with_multiple_sheets(data)
        else:
            properties = self.properties
            data = self._convert_data_to_dataframe(properties[self._EXPOSED_TYPE_PROPERTY], data)
            self._write_excel_with_single_sheet(
                data.to_excel, self._path, index=False, header=properties[self._HAS_HEADER_PROPERTY] or None
            )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass, field
from datetime import datetime
from functools import singledispatch
from typing import Any, Optional

from ..common._repr_enum import _ReprEnum
from ..exceptions.exceptions import InvalidEventAttributeName, InvalidEventOperation


class EventOperation(_ReprEnum):
    """Enum representing a type of operation performed on a Core entity.

    `EventOperation` is used as an attribute of the `Event^` object to describe the
    operation performed on an entity.<br>
    The possible operations are:

     - `CREATION`: Event related to a creation operation.
     - `UPDATE`: Event related to an update operation.
     - `DELETION`: Event related to a deletion operation.
     - `SUBMISSION`: Event related to a submission operation.

    """

    CREATION = 1
    UPDATE = 2
    DELETION = 3
    SUBMISSION = 4


class EventEntityType(_ReprEnum):
    """Enum representing an entity type.

    `EventEntityType` is used as an attribute of the `Event^` object to describe
    an entity that was changed.<br>
    The possible operations are:

    - `CYCLE`: Event related to a cycle entity.
    - `SCENARIO`: Event related to a scenario entity.
    - `SEQUENCE`: Event related to a sequence entity.
    - `TASK`: Event related to a task entity.
    - `DATA_NODE`: Event related to a data node entity.
    - `JOB`: Event related to a job entity.
    - `SUBMISSION`: Event related to a submission entity.
    """

    CYCLE = 1
    SCENARIO = 2
    SEQUENCE = 3
    TASK = 4
    DATA_NODE = 5
    JOB = 6
    SUBMISSION = 7


_NO_ATTRIBUTE_NAME_OPERATIONS = {EventOperation.CREATION, EventOperation.DELETION, EventOperation.SUBMISSION}
_UNSUBMITTABLE_ENTITY_TYPES = (
    EventEntityType.CYCLE,
    EventEntityType.DATA_NODE,
    EventEntityType.JOB,
    EventEntityType.SUBMISSION,
)
_ENTITY_TO_EVENT_ENTITY_TYPE = {
    "scenario": EventEntityType.SCENARIO,
    "sequence": EventEntityType.SEQUENCE,
    "task": EventEntityType.TASK,
    "data": EventEntityType.DATA_NODE,
    "job": EventEntityType.JOB,
    "cycle": EventEntityType.CYCLE,
    "submission": EventEntityType.SUBMISSION,
}


@dataclass(frozen=True)
class Event:
    """Event object used to notify any change in a Taipy application.

    An event holds the necessary attributes to identify the change.
    """

    entity_type: EventEntityType
    """Type of the entity that was changed (`DataNode^`, `Scenario^`, `Cycle^`, etc. )."""
    operation: EventOperation
    """Enum describing the operation that was performed on the entity.

    The operation is among `CREATION`, `UPDATE`, `DELETION`, and `SUBMISSION`.
    """
    entity_id: Optional[str] = None
    """Unique identifier of the entity that was changed."""
    attribute_name: Optional[str] = None
    """Name of the entity's attribute changed. Only relevant for `UPDATE` operations."""
    attribute_value: Optional[Any] = None
    """Value of the entity's attribute changed. Only relevant for `UPDATE` operations."""

    metadata: dict = field(default_factory=dict)
    """A dictionary of additional metadata about the source of this event."""
    creation_date: datetime = field(init=False)
    """Date and time of the event creation."""

    def __post_init__(self):
        # Creation date
        super().__setattr__("creation_date", datetime.now())

        # Check operation:
        if self.entity_type in _UNSUBMITTABLE_ENTITY_TYPES and self.operation == EventOperation.SUBMISSION:
            raise InvalidEventOperation

        # Check attribute name:
        if self.operation in _NO_ATTRIBUTE_NAME_OPERATIONS and self.attribute_name is not None:
            raise InvalidEventAttributeName


@singledispatch
def _make_event(
    entity: Any,
    operation: EventOperation,
    /,
    attribute_name: Optional[str] = None,
    attribute_value: Optional[Any] = None,
    **kwargs,
) -> Event:
    """Helper function to make an event for this entity with the given `EventOperation^` type.
    In case of `EventOperation.UPDATE^` events, an attribute name and value must be given.

    Parameters:
        entity (Any): The entity object to generate an event for.
        operation (EventOperation^): The operation of the event. The possible values are:
            <ul>
                <li>CREATION</li>
                <li>UPDATE</li>
                <li>DELETION</li>
                <li>SUBMISSION</li>
            </ul>
        attribute_name (Optional[str]): The name of the updated attribute for a `EventOperation.UPDATE`.
            This argument is always given in case of an UPDATE.
        attribute_value (Optional[Any]): The value of the updated attribute for a `EventOperation.UPDATE`.
            This argument is always given in case of an UPDATE.
        **kwargs (dict[str, any]): Any extra information that would be passed to the metadata event.
            Note: you should pass only simple types: str, float, double as values."""
    raise Exception(f"Unexpected entity type: {type(entity)}")



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from threading import Lock
from typing import List, Optional, Union

from taipy.common.logger._taipy_logger import _TaipyLogger

from .._entity._entity_ids import _EntityIds
from .._manager._manager import _Manager
from .._repository._abstract_repository import _AbstractRepository
from .._version._version_mixin import _VersionMixin
from ..exceptions.exceptions import SubmissionNotDeletedException
from ..job.job import Job, Status
from ..notification import EventEntityType, EventOperation, Notifier, _make_event
from ..reason import EntityDoesNotExist, ReasonCollection, SubmissionIsNotFinished
from ..scenario.scenario import Scenario
from ..sequence.sequence import Sequence
from ..submission.submission import Submission, SubmissionId, SubmissionStatus
from ..task.task import Task


class _SubmissionManager(_Manager[Submission], _VersionMixin):
    _ENTITY_NAME = Submission.__name__
    _repository: _AbstractRepository
    _EVENT_ENTITY_TYPE = EventEntityType.SUBMISSION
    __lock = Lock()
    __logger = _TaipyLogger._get_logger()

    @classmethod
    def _get_all(cls, version_number: Optional[str] = None) -> List[Submission]:
        """
        Returns all entities.
        """
        filters = cls._build_filters_with_version(version_number)
        return cls._repository._load_all(filters)

    @classmethod
    def _create(cls, entity_id: str, entity_type: str, entity_config: Optional[str], **properties) -> Submission:
        submission = Submission(
            entity_id=entity_id, entity_type=entity_type, entity_config_id=entity_config, properties=properties
        )
        cls._set(submission)

        Notifier.publish(_make_event(submission, EventOperation.CREATION))

        return submission

    @classmethod
    def _update_submission_status(cls, submission: Submission, job: Job) -> None:
        with cls.__lock:
            submission = cls._get(submission)

            if submission._submission_status == SubmissionStatus.FAILED:
                return

            job_status = job.status
            if job_status == Status.FAILED:
                cls.__set_submission_status(submission, SubmissionStatus.FAILED, job)
                cls.__logger.debug(
                    f"{job.id} status is {job_status}. Submission status set to `{submission._submission_status}`."
                )
                return
            if job_status == Status.CANCELED:
                submission._is_canceled = True
            elif job_status == Status.BLOCKED:
                submission._blocked_jobs.add(job.id)
                submission._pending_jobs.discard(job.id)
            elif job_status == Status.PENDING or job_status == Status.SUBMITTED:
                submission._pending_jobs.add(job.id)
                submission._blocked_jobs.discard(job.id)
            elif job_status == Status.RUNNING:
                submission._running_jobs.add(job.id)
                submission._pending_jobs.discard(job.id)
            elif job_status == Status.COMPLETED or job_status == Status.SKIPPED:
                submission._is_completed = True  # type: ignore
                submission._blocked_jobs.discard(job.id)
                submission._pending_jobs.discard(job.id)
                submission._running_jobs.discard(job.id)
            elif job_status == Status.ABANDONED:
                submission._is_abandoned = True  # type: ignore
                submission._running_jobs.discard(job.id)
                submission._blocked_jobs.discard(job.id)
                submission._pending_jobs.discard(job.id)
            cls._set(submission)

            # The submission_status is set later to make sure notification for updating
            # the submission_status attribute is triggered
            if submission._is_canceled:
                cls.__set_submission_status(submission, SubmissionStatus.CANCELED, job)
            elif submission._is_abandoned:
                cls.__set_submission_status(submission, SubmissionStatus.UNDEFINED, job)
            elif submission._running_jobs:
                cls.__set_submission_status(submission, SubmissionStatus.RUNNING, job)
            elif submission._pending_jobs:
                cls.__set_submission_status(submission, SubmissionStatus.PENDING, job)
            elif submission._blocked_jobs:
                cls.__set_submission_status(submission, SubmissionStatus.BLOCKED, job)
            elif submission._is_completed:
                cls.__set_submission_status(submission, SubmissionStatus.COMPLETED, job)
            else:
                cls.__set_submission_status(submission, SubmissionStatus.UNDEFINED, job)
            cls.__logger.debug(
                f"{job.id} status is {job_status}. Submission status set to `{submission._submission_status}`."
            )

    @classmethod
    def __set_submission_status(cls, submission: Submission, new_submission_status: SubmissionStatus, job: Job) -> None:
        if not submission._is_in_context:
            submission = cls._get(submission)
        _current_submission_status = submission._submission_status
        submission._submission_status = new_submission_status

        cls._set(submission)

        if _current_submission_status != submission._submission_status:
            event = _make_event(
                submission,
                EventOperation.UPDATE,
                "submission_status",
                submission._submission_status,
                job_triggered_submission_status_changed=job.id,
            )

            if not submission._is_in_context:
                Notifier.publish(event)
            else:
                submission._in_context_attributes_changed_collector.append(event)

    @classmethod
    def _get_latest(cls, entity: Union[Scenario, Sequence, Task]) -> Optional[Submission]:
        entity_id = entity.id if not isinstance(entity, str) else entity
        submissions_of_task = list(filter(lambda submission: submission.entity_id == entity_id, cls._get_all()))
        if len(submissions_of_task) == 0:
            return None
        if len(submissions_of_task) == 1:
            return submissions_of_task[0]
        else:
            return max(submissions_of_task)

    @classmethod
    def _delete(cls, submission: Union[Submission, SubmissionId]) -> None:
        if isinstance(submission, str):
            submission = cls._get(submission)
        if cls._is_deletable(submission):
            super()._delete(submission.id)
        else:
            err = SubmissionNotDeletedException(submission.id)
            cls._logger.error(err)
            raise err

    @classmethod
    def _hard_delete(cls, submission_id: SubmissionId) -> None:
        submission = cls._get(submission_id)
        entity_ids_to_delete = cls._get_children_entity_ids(submission)
        entity_ids_to_delete.submission_ids.add(submission.id)
        cls._delete_entities_of_multiple_types(entity_ids_to_delete)

    @classmethod
    def _get_children_entity_ids(cls, submission: Submission) -> _EntityIds:
        entity_ids = _EntityIds()

        for job in submission.jobs:
            entity_ids.job_ids.add(job.id)

        return entity_ids

    @classmethod
    def _is_deletable(cls, submission: Union[Submission, SubmissionId]) -> ReasonCollection:
        reason_collector = ReasonCollection()

        if isinstance(submission, str):
            submission_id = submission
            submission = cls._get(submission)
            if submission is None:
                reason_collector._add_reason(submission_id, EntityDoesNotExist(submission_id))
                return reason_collector

        if not submission.is_finished() and submission.submission_status != SubmissionStatus.UNDEFINED:
            reason_collector._add_reason(submission.id, SubmissionIsNotFinished(submission.id))

        return reason_collector



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
""" # Package for managing reasons why some Taipy operations are not allowed.

Reasons for the Taipy actions why they can't be performed.

Because Taipy applications are natively multiuser, asynchronous, and dynamic,
some functions should not be invoked in some specific contexts. You can protect
such calls by calling other methods that return a `ReasonCollection^`. It acts
like a Boolean value: True if the operation can be performed and False otherwise. If
the action cannot be performed, the `ReasonCollection^` holds all the individual
reasons as a list of `Reason^` objects.

This package exposes the `ReasonCollection^` class and all
the reason implementations that can be used to explain why some Taipy operations
are not available or allowed.
"""

from .reason import (
    DataNodeEditInProgress,
    DataNodeIsNotWritten,
    EntityDoesNotExist,
    EntityIsNotAScenario,
    EntityIsNotSubmittableEntity,
    InvalidUploadFile,
    JobIsNotFinished,
    NoFileToDownload,
    NotAFile,
    NotGlobalScope,
    Reason,
    ScenarioDoesNotBelongToACycle,
    ScenarioIsThePrimaryScenario,
    SubmissionIsNotFinished,
    SubmissionStatusIsUndefined,
    UploadFileCanNotBeRead,
    WrongConfigType,
)
from .reason_collection import ReasonCollection



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import multiprocessing as mp
from concurrent.futures import Executor, ProcessPoolExecutor
from functools import partial
from threading import Lock
from typing import Callable, Optional

from taipy.common.config import Config
from taipy.common.config._serializer._toml_serializer import _TomlSerializer

from ...job.job import Job
from .._abstract_orchestrator import _AbstractOrchestrator
from ._job_dispatcher import _JobDispatcher
from ._task_function_wrapper import _TaskFunctionWrapper


class _StandaloneJobDispatcher(_JobDispatcher):
    """Manages job dispatching (instances of `Job^` class) in an asynchronous way using a ProcessPoolExecutor."""

    _nb_available_workers_lock = Lock()
    _DEFAULT_MAX_NB_OF_WORKERS = 2

    def __init__(self, orchestrator: _AbstractOrchestrator, subproc_initializer: Optional[Callable] = None):
        super().__init__(orchestrator)
        max_workers = Config.job_config.max_nb_of_workers or self._DEFAULT_MAX_NB_OF_WORKERS
        self._executor: Executor = ProcessPoolExecutor(
            max_workers=max_workers, initializer=subproc_initializer, mp_context=mp.get_context("spawn")
        )
        self._nb_available_workers = self._executor._max_workers  # type: ignore

    def _can_execute(self) -> bool:
        """Returns True if the dispatcher have resources to dispatch a job."""
        with self._nb_available_workers_lock:
            self._logger.debug(f"{self._nb_available_workers=}")
            return self._nb_available_workers > 0

    def run(self):
        with self._executor:
            super().run()
        self._logger.debug("Standalone job dispatcher: Pool executor shut down.")

    def _dispatch(self, job: Job):
        """Dispatches the given `Job^` on an available worker for execution.

        Parameters:
            job (Job^): The job to submit on an executor with an available worker.
        """
        with self._nb_available_workers_lock:
            self._nb_available_workers -= 1
            self._logger.debug(f"Setting nb_available_workers to {self._nb_available_workers} in the dispatch method.")
        config_as_string = _TomlSerializer()._serialize(Config._applied_config)  # type: ignore[attr-defined]

        future = self._executor.submit(_TaskFunctionWrapper(job.id, job.task), config_as_string=config_as_string)
        future.add_done_callback(partial(self._update_job_status_from_future, job))

    def _update_job_status_from_future(self, job: Job, ft):
        with self._nb_available_workers_lock:
            self._nb_available_workers += 1
            self._logger.debug(f"Setting nb_available_workers to {self._nb_available_workers} in the callback method.")
        self._update_job_status(job, ft.result())



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .._repository._abstract_converter import _AbstractConverter
from ..common._utils import _load_fct
from ..data._data_manager_factory import _DataManagerFactory
from ..exceptions import NonExistingDataNode
from ..task._task_model import _TaskModel
from ..task.task import Task
from .task import TaskId


class _TaskConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, task: Task) -> _TaskModel:
        return _TaskModel(
            id=task.id,
            owner_id=task.owner_id,
            parent_ids=list(task._parent_ids),
            config_id=task.config_id,
            input_ids=cls.__to_ids(task.input.values()),
            function_name=task._function.__name__,
            function_module=task._function.__module__,
            output_ids=cls.__to_ids(task.output.values()),
            version=task._version,
            skippable=task._skippable,
            properties=task._properties.data.copy(),
        )

    @classmethod
    def _model_to_entity(cls, model: _TaskModel) -> Task:
        return Task(
            id=TaskId(model.id),
            owner_id=model.owner_id,
            parent_ids=set(model.parent_ids),
            config_id=model.config_id,
            function=_load_fct(model.function_module, model.function_name),
            input=cls.__to_data_nodes(model.input_ids),
            output=cls.__to_data_nodes(model.output_ids),
            version=model.version,
            skippable=model.skippable,
            properties=model.properties,
        )

    @staticmethod
    def __to_ids(data_nodes):
        return [i.id for i in data_nodes]

    @staticmethod
    def __to_data_nodes(data_nodes_ids):
        data_nodes = []
        data_manager = _DataManagerFactory._build_manager()
        for _id in data_nodes_ids:
            if data_node := data_manager._get(_id):
                data_nodes.append(data_node)
            else:
                raise NonExistingDataNode(_id)
        return data_nodes



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set

from taipy.common.config.common.scope import Scope

from .._version._version_manager_factory import _VersionManagerFactory
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit

in_memory_storage: Dict[str, Any] = {}


class InMemoryDataNode(DataNode):
    """Data Node stored in memory.

    Warning:
        This Data Node implementation is not compatible with a parallel execution of taipy tasks,
        but only with a task executor in development mode. The purpose of `InMemoryDataNode` is
        mostly to be used for development, prototyping, or debugging.

    The *properties* attribute can also contain the following optional entries:

    - *default_data* (`Any`): The default data of the data node. It is used at the data node
        instantiation
    """

    __STORAGE_TYPE = "in_memory"
    __DEFAULT_DATA_VALUE = "default_data"
    _REQUIRED_PROPERTIES: List[str] = []

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: List[Edit] = None,
        version: str = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        editor_id: Optional[str] = None,
        editor_expiration_date: Optional[datetime] = None,
        properties=None,
    ) -> None:
        if properties is None:
            properties = {}
        default_value = properties.pop(self.__DEFAULT_DATA_VALUE, None)
        super().__init__(
            config_id,
            scope,
            id,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            editor_id,
            editor_expiration_date,
            **properties,
        )
        if default_value is not None and self.id not in in_memory_storage:
            self._write(default_value)
            self._last_edit_date = datetime.now()
            self._edits.append(
                Edit(
                    {
                        "timestamp": self._last_edit_date,
                        "editor": "TAIPY",
                        "comment": "Default data written.",
                    }
                )
            )

        self._TAIPY_PROPERTIES.update({self.__DEFAULT_DATA_VALUE})

    @classmethod
    def storage_type(cls) -> str:
        """Return the storage type of the data node: "in_memory"."""
        return cls.__STORAGE_TYPE

    def _read(self):
        return in_memory_storage.get(self.id)

    def _write(self, data):
        in_memory_storage[self.id] = data



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
from importlib import util


def _check_dependency_is_installed(module_name: str, package_name: str) -> None:
    """
        Check if a package is installed.

        Args:
            module_name: Name of the taipy module importing the package.
            package_name: Name of the package.
    .
    """
    extras = {
        "boto3": "s3",
        "pymongo": "mongo",
    }
    if not util.find_spec(package_name):
        raise RuntimeError(
            f"Cannot use {module_name} as {package_name} package is not installed. Please install it  "
            f"using `pip install taipy[{extras.get(package_name)}]`."
        )


class EnterpriseEditionUtils:
    """NOT DOCUMENTED"""
    _TAIPY_ENTERPRISE_MODULE = "taipy.enterprise"
    _TAIPY_ENTERPRISE_CORE_MODULE = _TAIPY_ENTERPRISE_MODULE + ".core"
    _TAIPY_ENTERPRISE_EVENT_PACKAGE = _TAIPY_ENTERPRISE_MODULE + ".event"

    @classmethod
    def _using_enterprise(cls) -> bool:
        return util.find_spec(cls._TAIPY_ENTERPRISE_MODULE) is not None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import threading
from queue import Empty, SimpleQueue

from .event import Event


class _CoreEventConsumerBase(threading.Thread):
    """Abstract base class for implementing a Core event consumer.

    This class provides a framework for consuming events from a queue in a separate thread.
    It should be subclassed, and the `process_event` method should be implemented to define
    the custom logic for handling incoming events.

    ??? example "Basic usage"

        ```python
        class MyEventConsumer(CoreEventConsumerBase):
            def process_event(self, event: Event):
                # Custom event processing logic here
                print(f"Received event created at : {event.creation_date}")
                pass

        if __name__ == "__main__":
            registration_id, registered_queue = Notifier.register(
                entity_type=EventEntityType.SCENARIO,
                operation=EventOperation.CREATION
            )

            consumer = MyEventConsumer(registration_id, registered_queue)
            consumer.start()
            # ...
            consumer.stop()

            Notifier.unregister(registration_id)
        ```

        Firstly, we would create a consumer class extending from CoreEventConsumerBase
        and decide how to process the incoming events by defining the process_event.
        Then, we would specify the type of event we want to receive by registering with the Notifier.
        After that, we create an object of the consumer class by providing
        the registration_id and registered_queue and start consuming the event.
    """

    def __init__(self, registration_id: str, queue: SimpleQueue) -> None:
        """Initialize a CoreEventConsumerBase instance.

        Parameters:
            registration_id (str): A unique identifier of the registration. You can get a
                registration id invoking `Notifier.register()^` method.
            queue (SimpleQueue): The queue from which events will be consumed. You can get a
                queue invoking `Notifier.register()^` method.
        """
        threading.Thread.__init__(self, name=f"Thread-Taipy-Core-Consumer-{registration_id}")
        self.daemon = True
        self.queue = queue
        self.__STOP_FLAG = False
        self._TIMEOUT = 0.1

    def start(self) -> None:
        """Start the event consumer thread."""
        self.__STOP_FLAG = False
        threading.Thread.start(self)

    def stop(self) -> None:
        """Stop the event consumer thread."""
        self.__STOP_FLAG = True

    def run(self) -> None:
        while not self.__STOP_FLAG:
            try:
                event: Event = self.queue.get(block=True, timeout=self._TIMEOUT)
                self.process_event(event)
            except Empty:
                pass

    @abc.abstractmethod
    def process_event(self, event: Event) -> None:
        """This method should be overridden in subclasses to define how events are processed."""
        raise NotImplementedError



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import dataclasses
import enum
import json
from typing import Any, Dict

from ._decoder import _Decoder
from ._encoder import _Encoder


class _BaseModel:
    def __iter__(self):
        yield from self.__dict__.items()

    def to_dict(self) -> Dict[str, Any]:
        model_dict = {**dataclasses.asdict(self)}  # type: ignore[call-overload]

        for k, v in model_dict.items():
            if isinstance(v, enum.Enum):
                model_dict[k] = repr(v)
        return model_dict

    @staticmethod
    def _serialize_attribute(value):
        return json.dumps(value, ensure_ascii=False, cls=_Encoder)

    @staticmethod
    def _deserialize_attribute(value):
        if isinstance(value, str):
            return json.loads(value.replace("'", '"'), cls=_Decoder)
        return value

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        pass

    def to_list(self):
        pass



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Set, Union

import networkx as nx

from taipy.common.config.common._validate_id import _validate_id

from .._entity._entity import _Entity
from .._entity._labeled import _Labeled
from .._entity._properties import _Properties
from .._entity._reload import _Reloader, _self_reload, _self_setter
from .._entity.submittable import Submittable
from .._version._version_manager_factory import _VersionManagerFactory
from ..common._listattributes import _ListAttributes
from ..common._utils import _Subscriber
from ..data.data_node import DataNode
from ..exceptions.exceptions import AttributeKeyAlreadyExisted, NonExistingTask
from ..job.job import Job
from ..notification.event import Event, EventEntityType, EventOperation, _make_event
from ..submission.submission import Submission
from ..task.task import Task
from ..task.task_id import TaskId
from .sequence_id import SequenceId


class Sequence(_Entity, Submittable, _Labeled):
    """A subset of scenario tasks grouped to be executed together independently of the others.

    A sequence is attached to a `Scenario^`. It represents a subset of its tasks that need to
    be executed together, independently of the other tasks in the scenario. They must form a
    connected subgraph of the scenario's task graph. A scenario can hold multiple sequences.

    For instance, in a typical machine learning scenario, we may have several sequences:
    a sequence dedicated to preprocessing and preparing data, a sequence for computing a
    training model, and a sequence dedicated to scoring.

    ??? Example

        Let's assume we have a scenario configuration modelling a manufacturer that is
        training an ML model, predicting sales forecasts, and finally, based on
        the forecasts, planning its production. Three task are configured and linked
        together through data nodes.

        ![sequences](../../../../img/sequences.svg){ align=left }

        First, the sales sequence (boxed in green in the picture) contains **training**
        and **predict** tasks. Second, a production sequence (boxed in dark gray in the
        picture) contains the **planning** task.

        This problem has been modeled in two sequences - one sequence for the forecasting
        part and one for the production planning part. As a consequence, the two algorithms
        can have two different life cycles. They can run independently, under different
        schedules. For example, one on a fixed schedule (e.g. every week) and one on demand,
        interactively triggered by end-users.

        ```python
        import taipy as tp
        from taipy import Config

        def training(history):
            ...

        def predict(model, month):
            ...

        def planning(forecast, capacity):
            ...

        if __name__ == "__main__":
            # Configure data nodes
            sales_history_cfg = Config.configure_csv_data_node("sales_history")
            trained_model_cfg = Config.configure_data_node("trained_model")
            current_month_cfg = Config.configure_data_node("current_month")
            forecasts_cfg = Config.configure_data_node("sales_predictions")
            capacity_cfg = Config.configure_data_node("capacity")
            production_orders_cfg = Config.configure_sql_data_node("production_orders")

            # Configure tasks and scenarios
            train_cfg = Config.configure_task("train", function=training,
                                              input=sales_history_cfg, output=trained_model_cfg)
            predict_cfg = Config.configure_task("predict", function=predict,
                                                input=[trained_model_cfg, current_month_cfg],
                                                output=forecasts_cfg)
            plan_cfg = Config.configure_task("planning", function=planning,
                                            input=[forecasts_cfg, capacity_cfg],
                                            output=production_orders_cfg)
            scenario_cfg = Config.configure_scenario("scenario", task_configs=[train_cfg, predict_cfg, plan_cfg])

            # Create a new scenario and sequences
            scenario = tp.create_scenario(scenario_cfg)
            scenario.add_sequence("sales_sequence", [train_cfg, predict_cfg])
            scenario.add_sequence("production_sequence", [plan_cfg])

            # Get all sequences
            all_sequences = tp.get_sequences()

            # Submit one sequence only
            tp.submit(scenario.sales_sequence)
        ```

    Note that the sequences are not necessarily disjoint and may share some tasks.
    """

    _ID_PREFIX = "SEQUENCE"
    _SEPARATOR = "_"
    _MANAGER_NAME = "sequence"
    __CHECK_INIT_DONE_ATTR_NAME = "_init_done"

    id: SequenceId
    """The unique identifier of the sequence."""

    def __init__(
        self,
        properties: Dict[str, Any],
        tasks: Union[List[TaskId], List[Task], List[Union[TaskId, Task]]],
        sequence_id: SequenceId,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        subscribers: Optional[List[_Subscriber]] = None,
        version: Optional[str] = None,
    ):
        self.id: SequenceId = sequence_id
        super().__init__(self.id, subscribers)
        self._tasks = tasks
        self._owner_id = owner_id
        self._parent_ids = parent_ids or set()
        self._properties = _Properties(self, **properties)
        self._version = version or _VersionManagerFactory._build_manager()._get_latest_version()
        self._init_done = True

    def __hash__(self) -> int:
        """Return the hash of the sequence."""
        return hash(self.id)

    def __eq__(self, other) -> bool:
        """Check if a sequence is equal to another sequence."""
        return isinstance(other, Sequence) and self.id == other.id

    def __setattr__(self, name: str, value: Any) -> None:
        if self.__CHECK_INIT_DONE_ATTR_NAME not in dir(self) or name in dir(self):
            return super().__setattr__(name, value)
        else:
            try:
                self.__getattr__(name)
                raise AttributeKeyAlreadyExisted(name)
            except AttributeError:
                return super().__setattr__(name, value)

    def __getattr__(self, attribute_name: str):
        """Get the attribute of the sequence.

        The attribute can be a task or a data node.

        Parameters:
            attribute_name (str): The attribute name.

        Returns:
            The attribute value.

        Raises:
            AttributeError: If the attribute is not found.

        """
        protected_attribute_name = _validate_id(attribute_name)
        tasks = self._get_tasks()
        if protected_attribute_name in tasks:
            return tasks[protected_attribute_name]
        for task in tasks.values():
            if protected_attribute_name in task.input:
                return task.input[protected_attribute_name]
            if protected_attribute_name in task.output:
                return task.output[protected_attribute_name]
        raise AttributeError(f"{attribute_name} is not an attribute of sequence {self.id}")

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def tasks(self) -> Dict[str, Task]:
        """The dictionary of tasks used by the sequence."""
        return self._get_tasks()

    @tasks.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def tasks(self, tasks: Union[List[TaskId], List[Task]]) -> None:
        self._tasks = tasks

    @property
    def data_nodes(self) -> Dict[str, DataNode]:
        """The dictionary of data nodes used by the sequence."""
        data_nodes = {}
        list_data_nodes = [task.data_nodes for task in self._get_tasks().values()]
        for data_node in list_data_nodes:
            for k, v in data_node.items():
                data_nodes[k] = v
        return data_nodes

    @property
    def parent_ids(self) -> Set[str]:
        """The set of identifiers of the parent scenarios."""
        return self._parent_ids

    @property
    def owner_id(self) -> Optional[str]:
        """The identifier of the owner (scenario_id, cycle_id) or None."""
        return self._owner_id

    @property
    def version(self) -> str:
        """The application version of the sequence.

        The string indicates the application version of the sequence. If not
        provided, the latest version is used."""
        return self._version

    @property
    def properties(self) -> _Properties:
        """The dictionary of additional properties."""
        self._properties = _Reloader()._reload("sequence", self)._properties
        return self._properties

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def subscribers(self) -> _ListAttributes:
        """The list of callbacks to be called on `Job^`'s status change."""
        return self._subscribers

    @subscribers.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def subscribers(self, val) -> None:
        self._subscribers = _ListAttributes(self, val)

    def get_parents(self) -> Dict[str, Set[_Entity]]:
        """Get parent scenarios of the sequence.

        Returns:
            The dictionary of all parent entities.
                They are grouped by their type (Scenario^, Sequences^, or tasks^) so each key corresponds
                to a level of the parents and the value is a set of the parent entities.
                An empty dictionary is returned if the entity does not have parents.
        """
        from ... import core as tp

        return tp.get_parents(self)

    def subscribe(
        self,
        callback: Callable[[Sequence, Job], None],
        params: Optional[List[Any]] = None,
    ) -> None:
        """Subscribe a function to be called on `Job^` status change.
        The subscription is applied to all jobs created from the sequence's execution.

        Note:
            Notification will be available only for jobs created after this subscription.

        Parameters:
            callback (Callable[[Sequence^, Job^], None]): The callable function to be called on
                status change.
            params (Optional[List[Any]]): The parameters to be passed to the _callback_.
        """
        from ... import core as tp

        return tp.subscribe_sequence(callback, params, self)

    def unsubscribe(self, callback: Callable[[Sequence, Job], None], params: Optional[List[Any]] = None) -> None:
        """Unsubscribe a function that is called when the status of a `Job^` changes.

        Note:
            The function will continue to be called for ongoing jobs.

        Parameters:
            callback (Callable[[Sequence^, Job^], None]): The callable function to unsubscribe.
            params (Optional[List[Any]]): The parameters to be passed to the _callback_.
        """
        from ... import core as tp

        return tp.unsubscribe_sequence(callback, params, self)

    def submit(
        self,
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        **properties,
    ) -> Submission:
        """Submit the sequence for execution.

        All the `Task^`s of the sequence will be submitted for execution.

        Parameters:
            callbacks (List[Callable]): The list of callable functions to be called on status
                change.
            force (bool): Force execution even if the data nodes are in cache.
            wait (bool): Wait for the orchestrated jobs created from the sequence submission to be finished
                in asynchronous mode.
            timeout (Union[float, int]): The maximum number of seconds to wait for the jobs to be finished before
                returning.<br/>
                If not provided and *wait* is True, the function waits indefinitely.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            A `Submission^` containing the information of the submission.
        """
        from ._sequence_manager_factory import _SequenceManagerFactory

        return _SequenceManagerFactory._build_manager()._submit(self, callbacks, force, wait, timeout, **properties)

    def get_label(self) -> str:
        """Returns the sequence simple label prefixed by its owner label.

        Returns:
            The label of the sequence as a string.
        """
        return self._get_label()

    def get_simple_label(self) -> str:
        """Returns the sequence simple label.

        Returns:
            The simple label of the sequence as a string.
        """
        return self._get_simple_label()

    @staticmethod
    def _new_id(sequence_name: str, scenario_id) -> SequenceId:
        seq_id = sequence_name.replace(" ", "TPSPACE")
        return SequenceId(Sequence._SEPARATOR.join([Sequence._ID_PREFIX, _validate_id(seq_id), scenario_id]))

    def _is_consistent(self) -> bool:
        dag = self._build_dag()
        if dag.number_of_nodes() == 0:
            return True
        if not nx.is_directed_acyclic_graph(dag):
            return False
        if not nx.is_weakly_connected(dag):
            return False
        for left_node, right_node in dag.edges:
            if (isinstance(left_node, DataNode) and isinstance(right_node, Task)) or (
                isinstance(left_node, Task) and isinstance(right_node, DataNode)
            ):
                continue
            return False
        return True

    def _get_tasks(self) -> Dict[str, Task]:
        from ..task._task_manager_factory import _TaskManagerFactory

        tasks = {}
        task_manager = _TaskManagerFactory._build_manager()
        for task_or_id in self._tasks:
            t = task_manager._get(task_or_id, task_or_id)
            if not isinstance(t, Task):
                raise NonExistingTask(task_or_id)
            tasks[t.config_id] = t
        return tasks

    def _get_set_of_tasks(self) -> Set[Task]:
        from ..task._task_manager_factory import _TaskManagerFactory

        tasks = set()
        task_manager = _TaskManagerFactory._build_manager()
        for task_or_id in self._tasks:
            task = task_manager._get(task_or_id, task_or_id)
            if not isinstance(task, Task):
                raise NonExistingTask(task_or_id)
            tasks.add(task)
        return tasks


@_make_event.register(Sequence)
def _make_event_for_sequence(
    sequence: Sequence,
    operation: EventOperation,
    /,
    attribute_name: Optional[str] = None,
    attribute_value: Optional[Any] = None,
    **kwargs,
) -> Event:
    metadata = {**kwargs}
    return Event(
        entity_type=EventEntityType.SEQUENCE,
        entity_id=sequence.id,
        operation=operation,
        attribute_name=attribute_name,
        attribute_value=attribute_value,
        metadata=metadata,
    )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.common.logger._taipy_logger import _TaipyLogger

from .common._warnings import _warn_deprecated
from .orchestrator import Orchestrator


class Core:
    """NOT DOCUMENTED"""

    __logger = _TaipyLogger._get_logger()

    def __new__(cls) -> Orchestrator:  # type: ignore
        _warn_deprecated("'Core'", suggest="the 'Orchestrator' class")
        cls.__logger.warning(
            "The `Core` service is deprecated and replaced by the `Orchestrator` service. "
            "An `Orchestrator` instance has been instantiated instead."
        )
        return Orchestrator()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Dict

from taipy.common._cli._base_cli._abstract_cli import _AbstractCLI
from taipy.common._cli._base_cli._taipy_parser import _TaipyParser

from ..config import CoreSection


class _CoreCLI(_AbstractCLI):
    """Command-line interface for Taipy Core application."""

    _mode_args: Dict[str, Dict] = {
        "--development": {
            "action": "store_true",
            "dest": "taipy_development",
            "help": """
                When execute Taipy application in `development` mode, all entities from the previous development version
                will be deleted before running new Taipy application.
            """,
        },
        "--experiment": {
            "dest": "taipy_experiment",
            "nargs": "?",
            "const": "",
            "metavar": "VERSION",
            "help": """
                When execute Taipy application in `experiment` mode, the current Taipy application is saved to a new
                version. If version name already exists, check for compatibility with current Python Config and run the
                application. Without being specified, the version number will be a random string.
            """,
        },
    }

    __FORCE_ARGS: Dict[str, Dict] = {
        "--force": {
            "dest": "taipy_force",
            "action": "store_true",
            "help": """
                Force override the configuration of the version if existed and run the application. Default to False.
            """,
        },
        "--no-force": {
            "dest": "no_taipy_force",
            "action": "store_true",
            "help": "Stop the application if any Config conflict exists.",
        },
    }

    @classmethod
    def create_parser(cls):
        core_parser = _TaipyParser._add_groupparser("Taipy Core", "Optional arguments for Taipy Core service")

        mode_group = core_parser.add_mutually_exclusive_group()
        for mode_arg, mode_arg_dict in cls._mode_args.items():
            mode_group.add_argument(mode_arg, cls.__add_taipy_prefix(mode_arg), **mode_arg_dict)

        force_group = core_parser.add_mutually_exclusive_group()
        for force_arg, force_arg_dict in cls.__FORCE_ARGS.items():
            force_group.add_argument(cls.__add_taipy_prefix(force_arg), **force_arg_dict)

    @classmethod
    def create_run_parser(cls):
        run_parser = _TaipyParser._add_subparser("run", help="Run a Taipy application.")
        mode_group = run_parser.add_mutually_exclusive_group()
        for mode_arg, mode_arg_dict in cls._mode_args.items():
            mode_group.add_argument(mode_arg, **mode_arg_dict)

        force_group = run_parser.add_mutually_exclusive_group()
        for force_arg, force_arg_dict in cls.__FORCE_ARGS.items():
            force_group.add_argument(force_arg, **force_arg_dict)

    @classmethod
    def handle_command(cls):
        args, _ = _TaipyParser._parser.parse_known_args()

        as_dict = {}
        if args.taipy_development:
            as_dict[CoreSection._MODE_KEY] = "development"
        elif args.taipy_experiment is not None:
            as_dict[CoreSection._MODE_KEY] = "experiment"
            as_dict[CoreSection._VERSION_NUMBER_KEY] = args.taipy_experiment

        if args.taipy_force:
            as_dict[CoreSection._FORCE_KEY] = True
        elif args.no_taipy_force:
            as_dict[CoreSection._FORCE_KEY] = False

        return as_dict

    @classmethod
    def __add_taipy_prefix(cls, key: str):
        if key.startswith("--no-"):
            return f"{key[:5]}taipy-{key[5:]}"
        return f"{key[:2]}taipy-{key[2:]}"



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime

from taipy.common.config import Config

from .._repository._abstract_converter import _AbstractConverter
from .._version._version import _Version
from .._version._version_model import _VersionModel


class _VersionConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, version: _Version) -> _VersionModel:
        return _VersionModel(
            id=version.id, config=Config._to_json(version.config), creation_date=version.creation_date.isoformat()
        )  # type: ignore[attr-defined]

    @classmethod
    def _model_to_entity(cls, model: _VersionModel) -> _Version:
        version = _Version(id=model.id, config=Config._from_json(model.config))  # type: ignore[attr-defined]
        version.creation_date = datetime.fromisoformat(model.creation_date)
        return version



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import threading
import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Union

from .._entity._entity import _Entity
from .._entity._labeled import _Labeled
from .._entity._properties import _Properties
from .._entity._reload import _Reloader, _self_reload, _self_setter
from .._version._version_manager_factory import _VersionManagerFactory
from ..job.job import Job, JobId
from ..notification import Event, EventEntityType, EventOperation, _make_event
from ..reason.reason_collection import ReasonCollection
from .submission_id import SubmissionId
from .submission_status import SubmissionStatus


class Submission(_Entity, _Labeled):
    """Submission of a submittable entity: `Task^`, a `Sequence^` or a `Scenario^`.

    Task, Sequence, and Scenario entities can be submitted for execution. The submission
    represents the unique request to execute a submittable entity. The submission is created
    at the time the entity is submitted.

    The submission holds the jobs created by the execution of the submittable and the
    `SubmissionStatus^`. The status is lively updated by Taipy during the execution of the jobs.

    ??? example

        ```python
        import taipy as tp
        from taipy import Config

        def by_two(x: int):
            return x * 2

        if __name__ == "__main__":
            # Configure scenarios
            input_cfg = Config.configure_data_node("my_input")
            result_cfg = Config.configure_data_node("my_result")
            task_cfg = Config.configure_task("my_double", function=by_two, input=input_cfg, output=result_cfg)
            scenario_cfg = Config.configure_scenario("my_scenario", task_configs=[task_cfg])

            # Create a new scenario from the configuration
            scenario = tp.create_scenario(scenario_cfg)

            # Write the input data and submit the scenario
            scenario.my_input.write(3)
            submission = scenario.submit()

            # Retrieve the list of jobs, the submission status, and the creation date
            jobs = submission.jobs
            status = submission.submission_status
            creation_date = submission.creation_date
        ```
    """

    _ID_PREFIX = "SUBMISSION"
    _MANAGER_NAME = "submission"
    __SEPARATOR = "_"
    lock = threading.Lock()

    id: SubmissionId
    """The identifier of the `Submission` entity."""

    def __init__(
        self,
        entity_id: str,
        entity_type: str,
        entity_config_id: Optional[str] = None,
        id: Optional[SubmissionId] = None,
        jobs: Optional[Union[List[Job], List[JobId]]] = None,
        properties: Optional[Dict[str, Any]] = None,
        creation_date: Optional[datetime] = None,
        submission_status: Optional[SubmissionStatus] = None,
        version: Optional[str] = None,
    ):
        self._entity_id = entity_id
        self._entity_type = entity_type
        self._entity_config_id = entity_config_id
        self.id = id or self.__new_id()
        self._jobs: Union[List[Job], List[JobId], List] = jobs or []
        self._creation_date = creation_date or datetime.now()
        self._submission_status = submission_status or SubmissionStatus.SUBMITTED
        self._version = version or _VersionManagerFactory._build_manager()._get_latest_version()

        properties = properties or {}
        self._properties = _Properties(self, **properties.copy())

        self._is_abandoned = False
        self._is_completed = False
        self._is_canceled = False

        self._running_jobs: Set = set()
        self._blocked_jobs: Set = set()
        self._pending_jobs: Set = set()

    def __lt__(self, other) -> bool:
        """Compare the creation date of two submissions."""
        return self.creation_date.timestamp() < other.creation_date.timestamp()

    def __le__(self, other) -> bool:
        """Compare the creation date of two submissions."""
        return self.creation_date.timestamp() <= other.creation_date.timestamp()

    def __gt__(self, other) -> bool:
        """Compare the creation date of two submissions."""
        return self.creation_date.timestamp() > other.creation_date.timestamp()

    def __ge__(self, other) -> bool:
        """Compare the creation date of two submissions."""
        return self.creation_date.timestamp() >= other.creation_date.timestamp()

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other) -> bool:
        """Check if a submission is equal to another submission."""
        return isinstance(other, Submission) and self.id == other.id

    @property
    def entity_id(self) -> str:
        """The identifier of the entity that was submitted."""
        return self._entity_id

    @property
    def entity_type(self) -> str:
        """The type of the entity that was submitted."""
        return self._entity_type

    @property
    def entity_config_id(self) -> Optional[str]:
        """The config id of the entity that was submitted."""
        return self._entity_config_id

    @property
    def properties(self) -> _Properties:
        """A dictionary of additional properties."""
        self._properties = _Reloader()._reload(self._MANAGER_NAME, self)._properties
        return self._properties

    @property
    def creation_date(self) -> datetime:
        """The date and time when the submission was created."""
        return self._creation_date

    @property
    @_self_reload(_MANAGER_NAME)
    def submitted_at(self) -> Optional[datetime]:
        """The date and time when the submission was submitted.

        The submitted date and time corresponds to the date and time of the first job
        that was submitted. If no job was submitted, the submitted date and time is None.
        """
        jobs_submitted_at = [job.submitted_at for job in self.jobs if job.submitted_at]
        if jobs_submitted_at:
            return min(jobs_submitted_at)
        return None

    @property
    @_self_reload(_MANAGER_NAME)
    def run_at(self) -> Optional[datetime]:
        """The date and time when the submission was run.

        The run date and time corresponds to the date and time of the first job
        that was run. If no job was run, the run date and time is None.
        """
        jobs_run_at = [job.run_at for job in self.jobs if job.run_at]
        if jobs_run_at:
            return min(jobs_run_at)
        return None

    @property
    @_self_reload(_MANAGER_NAME)
    def finished_at(self) -> Optional[datetime]:
        """The date and time when the submission was finished.

        The finished date and time corresponds to the date and time of the last job
        that was completed. If at least one of the jobs is not finished, the finished
        date and time is None.
        """
        if all(job.finished_at for job in self.jobs):
            return max([job.finished_at for job in self.jobs if job.finished_at])
        return None

    @property
    @_self_reload(_MANAGER_NAME)
    def execution_duration(self) -> Optional[float]:
        """The duration of the submission execution in seconds.

        The execution duration in seconds is the duration from the first job running
        to the last job completion. If no job was run, the execution duration is None.
        If at least one job is not finished, the execution duration is the duration
        from the first job running time to the current time.
        """
        if self.finished_at and self.run_at:
            return (self.finished_at - self.run_at).total_seconds()
        elif self.run_at and self.finished_at is None:
            return (datetime.now() - self.run_at).total_seconds()
        return None

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def jobs(self) -> List[Job]:
        """The list of jobs created by the submission."""
        from ..job._job_manager_factory import _JobManagerFactory

        job_manager = _JobManagerFactory._build_manager()
        return [job_manager._get(job) for job in self._jobs]

    @jobs.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def jobs(self, jobs: Union[List[Job], List[JobId]]) -> None:
        self._jobs = jobs

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def submission_status(self) -> SubmissionStatus:
        """The status of the submission."""
        return self._submission_status

    @submission_status.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def submission_status(self, submission_status) -> None:
        self._submission_status = submission_status

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def is_abandoned(self) -> bool:
        """Indicate if the submission is abandoned."""
        return self._is_abandoned

    @is_abandoned.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def is_abandoned(self, val) -> None:
        self._is_abandoned = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def is_completed(self) -> bool:
        """Indicate if the submission is completed."""
        return self._is_completed

    @is_completed.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def is_completed(self, val) -> None:
        self._is_completed = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def is_canceled(self) -> bool:
        """Indicate if the submission is canceled."""
        return self._is_canceled

    @is_canceled.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def is_canceled(self, val) -> None:
        self._is_canceled = val

    @property
    def version(self) -> str:
        """The application version of the submission.

        The string indicates the application version of the submission. If not
        provided, the latest version is used."""
        return self._version

    def get_label(self) -> str:
        """Returns the submission simple label prefixed by its owner label.

        Returns:
            The label of the submission as a string.
        """
        return self._get_label()

    def get_simple_label(self) -> str:
        """Returns the submission simple label.

        Returns:
            The simple label of the submission as a string.
        """
        return self._get_simple_label()

    def is_finished(self) -> bool:
        """Indicate if the submission is finished.

        A submission is considered as finished if its submission status is
        `COMPLETED`, `FAILED`, or `CANCELED`.

        Returns:
            True if the submission is finished.
        """
        return self.submission_status in [
            SubmissionStatus.COMPLETED,
            SubmissionStatus.FAILED,
            SubmissionStatus.CANCELED,
        ]

    def is_deletable(self) -> ReasonCollection:
        """Indicate if the submission can be deleted.

        Returns:
            A ReasonCollection object that can function as a Boolean value,
                which is True if the submission can be deleted. False otherwise.
        """
        from ... import core as tp

        return tp.is_deletable(self)

    @staticmethod
    def __new_id() -> SubmissionId:
        """Generate a unique Submission identifier."""
        return SubmissionId(Submission.__SEPARATOR.join([Submission._ID_PREFIX, str(uuid.uuid4())]))


@_make_event.register(Submission)
def _make_event_for_submission(
    submission: Submission,
    operation: EventOperation,
    /,
    attribute_name: Optional[str] = None,
    attribute_value: Optional[Any] = None,
    **kwargs,
) -> Event:
    metadata = {
        "origin_entity_id": submission.entity_id,
        "origin_entity_type": submission.entity_type,
        "origin_entity_config_id": submission.entity_config_id,
        "creation_date": submission.creation_date,
        "version": submission._version,
        **kwargs,
    }
    return Event(
        entity_type=EventEntityType.SUBMISSION,
        entity_id=submission.id,
        operation=operation,
        attribute_name=attribute_name,
        attribute_value=attribute_value,
        metadata=metadata,
    )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime

from .._repository._abstract_converter import _AbstractConverter
from ..cycle._cycle_model import _CycleModel
from ..cycle.cycle import Cycle


class _CycleConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, cycle: Cycle) -> _CycleModel:
        return _CycleModel(
            id=cycle.id,
            name=cycle._name,
            frequency=cycle._frequency,
            creation_date=cycle._creation_date.isoformat(),
            start_date=cycle._start_date.isoformat(),
            end_date=cycle._end_date.isoformat(),
            properties=cycle._properties.data,
        )

    @classmethod
    def _model_to_entity(cls, model: _CycleModel) -> Cycle:
        return Cycle(
            id=model.id,
            name=model.name,
            frequency=model.frequency,
            properties=model.properties,
            creation_date=datetime.fromisoformat(model.creation_date),
            start_date=datetime.fromisoformat(model.start_date),
            end_date=datetime.fromisoformat(model.end_date),
        )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import TYPE_CHECKING, Dict, Set, Union

from ..notification import EventOperation, Notifier, _make_event
from ..reason import Reason, ReasonCollection

if TYPE_CHECKING:
    from ..data.data_node import DataNode, DataNodeId
    from ..scenario.scenario import Scenario, ScenarioId
    from ..sequence.sequence import Sequence, SequenceId
    from ..task.task import Task, TaskId


class _ReadyToRunProperty:
    IS_SUBMITTABLE_PROPERTY_NAME: str = "is_submittable"

    # A dictionary of the data nodes not ready_to_read and their corresponding submittable entities.
    _datanode_id_submittables: Dict["DataNodeId", Set[Union["ScenarioId", "SequenceId", "TaskId"]]] = {}

    # A nested dictionary of the submittable entities (Scenario, Sequence, Task) and
    # the data nodes that make it not ready_to_run with the reason(s)
    _submittable_id_datanodes: Dict[Union["ScenarioId", "SequenceId", "TaskId"], ReasonCollection] = {}

    @classmethod
    def _add(cls, dn: "DataNode", reason: Reason) -> None:
        from ..scenario.scenario import Scenario
        from ..sequence.sequence import Sequence
        from ..task.task import Task

        parent_entities = dn.get_parents()

        for scenario_parent in parent_entities.get(Scenario._MANAGER_NAME, []):
            if dn in scenario_parent.get_inputs():
                cls.__add(scenario_parent, dn, reason)  # type: ignore
        for sequence_parent in parent_entities.get(Sequence._MANAGER_NAME, []):
            if dn in sequence_parent.get_inputs():
                cls.__add(sequence_parent, dn, reason)  # type: ignore
        for task_parent in parent_entities.get(Task._MANAGER_NAME, []):
            if dn in task_parent.input.values():
                cls.__add(task_parent, dn, reason)  # type: ignore

    @classmethod
    def _remove(cls, datanode: "DataNode", reason: Reason) -> None:
        from ..taipy import get as tp_get

        # check the data node status to determine the reason to be removed
        submittable_ids: Set = cls._datanode_id_submittables.get(datanode.id, set())

        to_remove_dn = False
        for submittable_id in submittable_ids:
            # check remove the reason
            reason_entity = cls._submittable_id_datanodes.get(submittable_id)
            if reason_entity is not None:
                reason_entity._remove_reason(datanode.id, reason)
                to_remove_dn = not reason_entity._entity_id_exists_in_reason(datanode.id)
                if reason_entity:
                    submittable = tp_get(submittable_id)
                    cls.__publish_submittable_property_event(submittable, True)
                    cls._submittable_id_datanodes.pop(submittable_id)

        if to_remove_dn:
            cls._datanode_id_submittables.pop(datanode.id)

    @classmethod
    def __add(cls, submittable: Union["Scenario", "Sequence", "Task"], datanode: "DataNode", reason: Reason) -> None:
        if datanode.id not in cls._datanode_id_submittables:
            cls._datanode_id_submittables[datanode.id] = set()
        cls._datanode_id_submittables[datanode.id].add(submittable.id)

        if submittable.id not in cls._submittable_id_datanodes:
            cls.__publish_submittable_property_event(submittable, False)

        if submittable.id not in cls._submittable_id_datanodes:
            cls._submittable_id_datanodes[submittable.id] = ReasonCollection()
        cls._submittable_id_datanodes[submittable.id]._add_reason(datanode.id, reason)

    @staticmethod
    def __publish_submittable_property_event(
        submittable: Union["Scenario", "Sequence", "Task"], submittable_property
    ) -> None:
        Notifier.publish(
            _make_event(
                submittable,
                EventOperation.UPDATE,
                attribute_name=_ReadyToRunProperty.IS_SUBMITTABLE_PROPERTY_NAME,
                attribute_value=submittable_property,
            )
        )
